/*
   Parallel search for "aardvark" in the files in the current directory.
   Each file is searched in parallel.

   This version does its own buffering of I/O from a FileInputStream.
*/

import java.time.Instant;
import java.time.Duration;
import java.util.Date;
import java.text.SimpleDateFormat;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.io.IOException;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.nio.file.Path;
import java.util.*;

public class SearchParallel_v7_Files
{
   public static final int cores = Runtime.getRuntime().availableProcessors();
   public static final ExecutorService pool = Executors.newFixedThreadPool(cores);

   public static int totalLines = 0; // There is a race condition on this variable!

   public static void main(String[] args) throws IOException
   {
      final SimpleDateFormat ft = new SimpleDateFormat("hh:mm:ss");
      System.out.println(ft.format(new Date()) + " Search Parallel ver 7 in all Files");
      Instant start = Instant.now();

      Files.list(Paths.get("."))
             .filter(p -> p.toString().endsWith(".txt"))
               .forEach(p -> pool.execute(new Task2(p))); // files are processed in parallel

      Instant stop = Instant.now();
      long time = Duration.between(start, stop).toMillis();
      System.out.printf("%s All files submitted to the pool.\n", ft.format(new Date()));

    //pool.shutdown(); // Can't shut down here because tasks in pool are submitting new tasks.

      // Wait for the work queue to be empty.
      while ( ! ((ThreadPoolExecutor)pool).getQueue().isEmpty() )
      {
         try{Thread.sleep(10);}catch(Exception e){}
      }

      stop = Instant.now();
      time = Duration.between(start, stop).toMillis();
      System.out.printf("%s processed %,d lines from all files.\n",
                       ft.format(new Date()), totalLines);
      System.out.printf("elapsed time: %,dms\n", time);

      pool.shutdown();
      try
      {
         pool.awaitTermination(5, TimeUnit.MINUTES);
      }
      catch (InterruptedException e)
      {
         pool.shutdownNow();
         Thread.currentThread().interrupt(); // Preserve interrupt status
      }
      System.out.println( pool.toString() );
   }
}//main()


// A Task2 object, when it is called by one of the threads in
// the Executor's thread pool, will search one file of text.
class Task2 implements Runnable
{
   private static final SimpleDateFormat ft = new SimpleDateFormat("hh:mm:ss");
   private static final int BUFFERSIZE = 8 * 8192;

   private Path path;

   public Task2(Path path)
   {
      this.path = path;
   }

   public void run()
   {
      byte[] c = new byte[BUFFERSIZE];
      int n;
      int lineNumber = 1;

      try
      {
         // Open the file.
         final String fileName = path.getFileName().toString();
         final FileInputStream fis = new FileInputStream( fileName );

         while ( (n = fis.read(c)) != -1 )
         {
            SearchParallel_v7_Files.pool.execute( new Task(c, n, lineNumber, fileName) );
            for (int i = 0; i < n; i++)
            {
               if ( '\n' == c[i] )
               {
                  ++lineNumber;
               }
            }
            c = new byte[BUFFERSIZE];
         }
         fis.close();
         SearchParallel_v7_Files.totalLines += lineNumber; // There is a race condition on totalLines!
      }
      catch(IOException e){throw new Error(e);}
   }
}//Task2


// A Task object, when it is called by one of the threads in
// the Executor's thread pool, will filter one line of text.
class Task implements Runnable
{
   private static final SimpleDateFormat ft = new SimpleDateFormat("hh:mm:ss");
   private final byte[] buffer;
   private final int n;
   private int lineNumber;
   private final String fileName;

   public Task(final byte[] buffer, final int n, final int lineNumber, final String fileName)
   {
      this.buffer = buffer;
      this.n = n;
      this.lineNumber = lineNumber;
      this.fileName = fileName;
   }

   public void run()
   {
      // Filter the data in the buffer.
      for (int i = 0; i < n; i++)
      {
         if ( '\n' == buffer[i] )
         {
            ++lineNumber;
         }
         else if ( i < n - 8
             && 'a' == buffer[i]
                && 'a' == buffer[i+1]
                && 'r' == buffer[i+2]
                && 'd' == buffer[i+3]
                && 'v' == buffer[i+4]
                && 'a' == buffer[i+5]
                && 'r' == buffer[i+6]
                && 'k' == buffer[i+7] )
         {
            System.out.printf("%s line %,d in %s (total line no %,d).\n",
                            ft.format(new Date()),
                               lineNumber,
                               fileName,
                               lineNumber + SearchParallel_v7_Files.totalLines);
         }
      }
   }
}//Task
